﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.EvolutionTracking;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Tracking.Analyzers {
  [Item("SymbolicDataAnalysisFragmentLengthAnalyzer", "An analyzer for fragment length.")]
  [StorableClass]
  public class SymbolicDataAnalysisFragmentLengthAnalyzer : EvolutionTrackingAnalyzer<ISymbolicExpressionTree> {
    [StorableConstructor]
    protected SymbolicDataAnalysisFragmentLengthAnalyzer(bool deserializing) : base(deserializing) { }

    public SymbolicDataAnalysisFragmentLengthAnalyzer() {
      UpdateCounterParameter.ActualName = "FragmentLengthAnalyzerUpdateCounter";
    }

    protected SymbolicDataAnalysisFragmentLengthAnalyzer(SymbolicDataAnalysisFragmentLengthAnalyzer original, Cloner cloner) : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) { return new SymbolicDataAnalysisFragmentLengthAnalyzer(this, cloner); }

    public override IOperation Apply() {
      int updateInterval = UpdateIntervalParameter.Value.Value;
      IntValue updateCounter = UpdateCounterParameter.ActualValue;
      // if counter does not yet exist then initialize it with update interval
      // to make sure the solutions are analyzed on the first application of this operator
      if (updateCounter == null) {
        updateCounter = new IntValue(updateInterval);
        UpdateCounterParameter.ActualValue = updateCounter;
      }
      //analyze solutions only every 'updateInterval' times
      if (updateCounter.Value != updateInterval) {
        updateCounter.Value++;
        return base.Apply();
      }
      updateCounter.Value = 1;

      if (PopulationGraph == null || Generation.Value == 0)
        return base.Apply();

      // consider all fragments (including those of the intermediate crossover children)
      double averageCrossoverFragmentLength = 0;
      double averageMutationFragmentLength = 0;
      int crossoverChildren = 0;
      int mutationChildren = 0;

      var vertices = PopulationGraph.Vertices.Where(x => x.Rank > Generation.Value - 1);
      foreach (var v in vertices) {
        if (!v.InArcs.Any() || v.InArcs.Last().Data == null)
          continue;
        var fragment = (IFragment<ISymbolicExpressionTreeNode>)v.InArcs.Last().Data;
        if (v.InDegree == 2) {
          averageCrossoverFragmentLength += fragment.Root.GetLength();
          crossoverChildren++;
        } else {
          averageMutationFragmentLength += fragment.Root.GetLength();
          mutationChildren++;
        }
      }
      var averageFragmentLength = (averageCrossoverFragmentLength + averageMutationFragmentLength) / (crossoverChildren + mutationChildren);
      averageCrossoverFragmentLength /= crossoverChildren;
      averageMutationFragmentLength /= mutationChildren;

      DataTable table;
      if (!Results.ContainsKey("AverageFragmentLength")) {
        table = new DataTable("Average fragment length");
        var row = new DataRow("Average fragment length") { VisualProperties = { StartIndexZero = true } };
        row.Values.Add(averageFragmentLength);
        table.Rows.Add(row);
        row = new DataRow("Average crossover fragment length") { VisualProperties = { StartIndexZero = true } };
        row.Values.Add(averageCrossoverFragmentLength);
        table.Rows.Add(row);
        row = new DataRow("Average mutation fragment length") { VisualProperties = { StartIndexZero = true } };
        row.Values.Add(averageMutationFragmentLength);
        table.Rows.Add(row);
        Results.Add(new Result("AverageFragmentLength", table));
      } else {
        table = (DataTable)Results["AverageFragmentLength"].Value;
        table.Rows["Average fragment length"].Values.Add(averageFragmentLength);
        table.Rows["Average crossover fragment length"].Values.Add(averageCrossoverFragmentLength);
        table.Rows["Average mutation fragment length"].Values.Add(averageMutationFragmentLength);
      }
      return base.Apply();
    }
  }
}
